import numpy as np
cimport numpy as np
cimport cython

cdef extern from "Python.h":
    ctypedef int Py_intptr_t

cdef extern from "numpy/arrayobject.h":
    ctypedef class numpy.ndarray [object PyArrayObject]:
        cdef char *data
        cdef Py_intptr_t *dimensions
        cdef Py_intptr_t *shape        
        cdef Py_intptr_t *strides
    cdef void import_array()
    cdef int  PyArray_ITEMSIZE(np.ndarray)

import_array()

@cython.boundscheck(False)
def skeletonize_loop(np.ndarray[dtype=np.uint8_t, ndim=2, 
                                negative_indices=False, mode='c'] result,
                     np.ndarray[dtype=np.int32_t, ndim=1,
                                negative_indices=False, mode='c'] i,
                     np.ndarray[dtype=np.int32_t, ndim=1,
                                negative_indices=False, mode='c'] j,
                     np.ndarray[dtype=np.int32_t, ndim=1,
                                negative_indices=False, mode='c'] order,
                     np.ndarray[dtype=np.uint8_t, ndim=1,
                                negative_indices=False, mode='c'] table):
    '''Inner loop of skeletonize function

    result - on input, the image to be skeletonized, on output the skeletonized
             image
    i,j    - the coordinates of each foreground pixel in the image
    order  - the index of each pixel, in the order of processing
    table  - the 512-element lookup table of values after transformation

    The loop determines whether each pixel in the image can be removed without
    changing the Euler number of the image. The pixels are ordered by
    increasing distance from the background which means a point nearer to
    the quench-line of the brushfire will be evaluated later than a
    point closer to the edge.
    '''
    cdef:
        np.int32_t accumulator
        np.int32_t index,order_index
        np.int32_t ii,jj

    for index in range(order.shape[0]):
        accumulator = 16
        order_index = order[index]
        ii = i[order_index]
        jj = j[order_index]
        if ii > 0:
            if jj > 0 and result[ii-1,jj-1]:
                accumulator += 1
            if result[ii-1,jj]:
                accumulator += 2
            if jj < result.shape[1]-1 and result[ii-1,jj+1]:
                    accumulator += 4
            if jj > 0 and result[ii,jj-1]:
                accumulator += 8
            if jj < result.shape[1]-1 and result[ii,jj+1]:
                accumulator += 32
            if ii < result.shape[0]-1:
                if jj > 0 and result[ii+1,jj-1]:
                    accumulator += 64
                if result[ii+1,jj]:
                    accumulator += 128
                if jj < result.shape[1]-1 and result[ii+1,jj+1]:
                    accumulator += 256
            result[ii,jj] = table[accumulator]

@cython.boundscheck(False)
def table_lookup_index(np.ndarray[dtype=np.uint8_t, ndim=2,
                                  negative_indices=False, mode='c'] image):
    '''Return an index into a table per pixel of a binary image

    Take the sum of true neighborhood pixel values where the neighborhood
    looks like this:
     1   2   4
     8  16  32
    64 128 256

    This code could be replaced by a convolution with the kernel:
    256 128 64
     32  16  8
      4   2  1
    but this runs about twice as fast because of inlining and the
    hardwired kernel.
    '''
    cdef:
        np.ndarray[dtype=np.int32_t, ndim=2, 
                   negative_indices=False, mode='c'] indexer
        np.int32_t *p_indexer
        np.uint8_t *p_image
        np.int32_t i_stride
        np.int32_t i_shape
        np.int32_t j_shape
        np.int32_t i
        np.int32_t j
        np.int32_t offset

    i_shape   = image.shape[0]
    j_shape   = image.shape[1]
    indexer = np.zeros((i_shape,j_shape),np.int32)
    p_indexer = <np.int32_t *>indexer.data
    p_image   = <np.uint8_t *>image.data
    i_stride  = image.strides[0]
    assert i_shape >= 3 and j_shape >= 3, "Please use the slow method for arrays < 3x3"

    for i in range(1,i_shape-1):
        offset = i_stride*i+1
        for j in range(1,j_shape-1):
            if p_image[offset]:
                p_indexer[offset+i_stride+1] += 1
                p_indexer[offset+i_stride] += 2
                p_indexer[offset+i_stride-1] += 4
                p_indexer[offset+1] += 8
                p_indexer[offset] += 16
                p_indexer[offset-1] += 32
                p_indexer[offset-i_stride+1] += 64
                p_indexer[offset-i_stride] += 128
                p_indexer[offset-i_stride-1] += 256
            offset += 1
    #
    # Do the corner cases (literally)
    #
    if image[0,0]:
        indexer[0,0] += 16
        indexer[0,1] += 8
        indexer[1,0] += 2
        indexer[1,1] += 1

    if image[0,j_shape-1]:
        indexer[0,j_shape-2] += 32
        indexer[0,j_shape-1] += 16
        indexer[1,j_shape-2] += 4
        indexer[1,j_shape-1] += 2

    if image[i_shape-1,0]:
        indexer[i_shape-2,0] += 128
        indexer[i_shape-2,1] += 64
        indexer[i_shape-1,0] += 16
        indexer[i_shape-1,1] += 8

    if image[i_shape-1,j_shape-1]:
        indexer[i_shape-2,j_shape-2] += 256
        indexer[i_shape-2,j_shape-1] += 128
        indexer[i_shape-1,j_shape-2] += 32
        indexer[i_shape-1,j_shape-1] += 16
    #
    # Do the edges
    #
    for j in range(1,j_shape-1):
        if image[0,j]:
            indexer[0,j-1] += 32
            indexer[0,j]   += 16
            indexer[0,j+1] += 8
            indexer[1,j-1] += 4
            indexer[1,j]   += 2
            indexer[1,j+1] += 1
        if image[i_shape-1,j]:
            indexer[i_shape-2,j-1] += 256
            indexer[i_shape-2,j]   += 128
            indexer[i_shape-2,j+1] += 64
            indexer[i_shape-1,j-1] += 32
            indexer[i_shape-1,j]   += 16
            indexer[i_shape-1,j+1] += 8

    for i in range(1,i_shape-1):
        if image[i,0]:
            indexer[i-1,0] += 128
            indexer[i,0]   += 16
            indexer[i+1,0] += 2
            indexer[i-1,1] += 64
            indexer[i,1]   += 8
            indexer[i+1,1] += 1
        if image[i,j_shape-1]:
            indexer[i-1,j_shape-2] += 256
            indexer[i,j_shape-2]   += 32
            indexer[i+1,j_shape-2] += 4
            indexer[i-1,j_shape-1] += 128
            indexer[i,j_shape-1]   += 16
            indexer[i+1,j_shape-1] += 2
    return indexer

@cython.boundscheck(False)
def grey_reconstruction_loop(np.ndarray[dtype=np.uint32_t, ndim=1,
                                        negative_indices = False,
                                        mode = 'c'] avalues,
                             np.ndarray[dtype=np.int32_t, ndim=1,
                                        negative_indices = False,
                                        mode = 'c'] aprev,
                             np.ndarray[dtype=np.int32_t, ndim=1,
                                        negative_indices = False,
                                        mode = 'c'] anext,
                             np.ndarray[dtype=np.int32_t, ndim=1,
                                        negative_indices = False,
                                        mode = 'c'] astrides,
                             np.int32_t current,
                             int image_stride):
    '''The inner loop for grey_reconstruction'''
    cdef:
        np.int32_t neighbor
        np.uint32_t neighbor_value
        np.uint32_t current_value
        np.uint32_t mask_value
        np.int32_t link
        int i
        np.int32_t nprev
        np.int32_t nnext
        int nstrides = astrides.shape[0]
        np.uint32_t *values = <np.uint32_t *>(avalues.data)
        np.int32_t *prev = <np.int32_t *>(aprev.data)
        np.int32_t *next = <np.int32_t *>(anext.data)
        np.int32_t *strides = <np.int32_t *>(astrides.data)
    
    while current != -1:
        if current < image_stride:
            current_value = values[current]
            if current_value == 0:
                break
            for i in range(nstrides):
                neighbor = current + strides[i]
                neighbor_value = values[neighbor]
                # Only do neighbors less than the current value
                if neighbor_value < current_value:
                    mask_value = values[neighbor + image_stride]
                    # Only do neighbors less than the mask value
                    if neighbor_value < mask_value:
                        # Raise the neighbor to the mask value if
                        # the mask is less than current
                        if mask_value < current_value:
                            link = neighbor + image_stride
                            values[neighbor] = mask_value
                        else:
                            link = current
                            values[neighbor] = current_value
                        # unlink the neighbor
                        nprev = prev[neighbor]
                        nnext = next[neighbor]
                        next[nprev] = nnext
                        if nnext != -1:
                            prev[nnext] = nprev
                        # link the neighbor after the link
                        nnext = next[link]
                        next[neighbor] = nnext
                        prev[neighbor] = link
                        if nnext >= 0:
                            prev[nnext] = neighbor
                            next[link] = neighbor
        current = next[current]
        
@cython.boundscheck(False)
def _all_connected_components(np.ndarray[dtype=np.uint32_t, ndim=1,
                                         negative_indices = False,
                                         mode = 'c'] i_a,
                              np.ndarray[dtype=np.uint32_t, ndim=1,
                                         negative_indices = False,
                                         mode = 'c'] j_a,
                              np.ndarray[dtype=np.uint32_t, ndim=1,
                                         negative_indices = False,
                                         mode = 'c'] indexes_a,
                              np.ndarray[dtype=np.uint32_t, ndim=1,
                                         negative_indices = False,
                                         mode = 'c'] counts_a,
                              np.ndarray[dtype=np.uint32_t, ndim=1,
                                         negative_indices = False,
                                         mode = 'c'] label_a):
    '''Inner loop for the all_connected_components algorithm
    
    i,j - vectors giving the edges between vertices. These vectors must
          be pre-sorted by i then j. There should be one j,i edge for
          every i,j edge for all_connected_components.
          
    indexes - 1 index per vertex # into the i,j array giving the index
              of the first edge for vertex i
              
    counts - 1 count per vertex # of the # of edges for vertex i
    
    label - one array element per vertex, the label assigned to this
            vertex's connected components. On exit, this is the vertex
            number of the first vertex processed in the connected component.
    '''
    cdef:
        # n = # of vertices
        np.uint32_t n = counts_a.shape[0]
        np.ndarray[dtype=np.uint8_t, ndim=1,
                   negative_indices = False,
                   mode = 'c'] being_processed_a = np.zeros(n, np.uint8)
        np.ndarray[dtype=np.uint32_t, ndim=1,
                   negative_indices = False,
                   mode = 'c'] visit_index_a = np.zeros(n, np.uint32)
        np.ndarray[dtype=np.uint32_t, ndim=1,
                   negative_indices = False,
                   mode = 'c'] stack_v_a = np.zeros(n, np.uint32)
        np.ndarray[dtype=np.uint32_t, ndim=1,
                   negative_indices = False,
                   mode = 'c'] v_idx_a  = np.zeros(n, np.uint32)
        np.ndarray[dtype=np.uint32_t, ndim=1,
                   negative_indices = False,
                   mode = 'c'] stack_being_processed_a = np.zeros(n, np.uint32)
        #
        # This is the recursion depth for recursive calls to process the
        # connected components.
        #
        np.uint32_t  stack_ptr = 0
        #
        # cur_index gives the next visit_index to be assigned. The visit_index
        # orders vertices according to when they were first visited
        #
        np.uint32_t  cur_index = 0
        #
        # raw pointer to the "from" vertices in the edge list (unused)
        #
        np.uint32_t *i = <np.uint32_t *>(i_a.data)
        #
        # raw pointer to the "to" vertices in the edge list
        #
        np.uint32_t *j = <np.uint32_t *>(j_a.data)
        #
        # raw pointer to the index into j for each vertex
        #
        np.uint32_t *indexes = <np.uint32_t *>(indexes_a.data)
        #
        # raw pointer to the # of edges per vertex
        #
        np.uint32_t *counts = <np.uint32_t *>(counts_a.data)
        #
        # the label per vertex
        #
        np.uint32_t *label = <np.uint32_t *>(label_a.data)
        #
        # The recursive "function" has a single argument - the vertex to be
        # processed, so this mimics a stack where v is pushed on entry
        # and popped on exit.
        #
        np.uint32_t *stack_v = <np.uint32_t *>(stack_v_a.data)
        #
        # The recursive function has a single local variable
        # v_idx which is the index of the current edge being processed.
        # On entry, this is UNDEFINED to indicate that the recursive function
        # should initialize "v". Afterwards, it cycles from 0 to the count.
        # When it reaches the count, the recursive function "exits"
        #
        np.uint32_t *v_idx = <np.uint32_t *>(v_idx_a.data)
        #
        # Holds the index of the top-level v in the top-level loop
        #
        np.uint32_t v
        #
        # Holds the v inside the recursive function
        #
        np.uint32_t vv
        #
        # Holds the vertex at the opposite side of the edge from vv
        #
        np.uint32_t v1
        #
        # A value to indicate something that's not been processed
        #
        np.uint32_t UNDEFINED = -1
        
    label_a[:] = UNDEFINED
    v_idx_a[:] = UNDEFINED
        
    for v in range(n):
        if label[v] == UNDEFINED:
            stack_v[0] = v
            stack_ptr = 1
            while(stack_ptr > 0):
                vv = stack_v[stack_ptr-1]
                if v_idx[vv] == UNDEFINED:
                    #
                    # Start of recursive function for a vertex: set the vertex
                    #     label.
                    #
                    label[vv] = cur_index
                    v_idx[vv] = 0
                if v_idx[vv] < counts[vv]:
                    # For each edge of v, push other vertex on stack 
                    # if unprocessed.
                    #
                    v1 = j[indexes[vv] + v_idx[vv]]
                    v_idx[vv] += 1
                    if label[v1] == UNDEFINED:
                        stack_v[stack_ptr] = v1
                        stack_ptr += 1
                else:
                    #
                    # We have processed every edge for vv.
                    #
                    stack_ptr -= 1
            cur_index += 1                                    

@cython.boundscheck(False)
def index_lookup(np.ndarray[dtype=np.int32_t, ndim=1, negative_indices=False] index_i, 
                 np.ndarray[dtype=np.int32_t, ndim=1, negative_indices=False] index_j, 
                 np.ndarray[dtype=np.uint32_t, ndim=2, negative_indices=False] image,
                 table_in, 
                 iterations=None):
    '''Perform a table lookup for only the indexed pixels
    
    For morphological operations that only convert 1 to 0, the set of
    resulting pixels is always a subset of the input set. Therefore, when
    repeating, it will be faster to operate only on the subsets especially
    when the results are 1-d or 0-d objects.
    
    This function returns a new index_i and index_j array of the pixels
    that survive the operation. The image is modified in-place to remove
    the pixels that did not survive.
    
    index_i - an array of row indexes into the image.
    index_j - a similarly-shaped array of column indexes.
    image - the binary image: *NOTE* add a row and column of border values
            to the original image to account for pixels on the edge of the
            image.
    iterations - # of iterations to do, default is "forever"
    
    The idea of index_lookup was taken from
    http://blogs.mathworks.com/steve/2008/06/13/performance-optimization-for-applylut/
    which, apparently, is how Matlab achieved its bwmorph speedup.
    '''
    cdef:
        np.ndarray[dtype=np.uint8_t, ndim=1, negative_indices=False] table = table_in.astype(np.uint8)
        np.uint32_t center, hit_count, idx, indexer
        np.int32_t idxi, idxj

    if iterations == None:
        # Worst case - remove one per iteration
        iterations = len(index_i)

    for i in range(iterations):
        hit_count = len(index_i)
        with nogil:
            #
            # For integer images (i.e., labels), a neighbor point is
            # "background" if it doesn't match the central value. This
            # lets adjacent labeled objects shrink independently of each
            # other.
            #
            for 0 <= idx < hit_count:
                idxi, idxj = index_i[idx], index_j[idx]
                center = image[idxi, idxj]
                indexer =          ((image[idxi - 1, idxj - 1] == center) * 1 +
                                    (image[idxi - 1, idxj]     == center) * 2 +
                                    (image[idxi - 1, idxj + 1] == center) * 4 +
                                    (image[idxi,     idxj - 1] == center) * 8 +
                                    16 +
                                    (image[idxi,     idxj + 1] == center) * 32 +
                                    (image[idxi + 1, idxj - 1] == center) * 64 +
                                    (image[idxi + 1, idxj]     == center) * 128 +
                                    (image[idxi + 1, idxj + 1] == center) * 256)
                if table[indexer] == 0:
                    # mark for deletion
                    index_i[idx] = -index_i[idx]

            # remove marked pixels
            for 0 <= idx < hit_count:
                idxi, idxj = index_i[idx], index_j[idx]
                if idxi < 0:
                    image[-idxi, idxj] = 0

        index_j = index_j[index_i >= 0]
        index_i = index_i[index_i >= 0]
        if len(index_i) == hit_count:
            break

    return (index_i, index_j)

def prepare_for_index_lookup(image, border_value):
    '''Return the index arrays of "1" pixels and an image with an added border
    
    The routine, index_lookup takes an array of i indexes, an array of
    j indexes and an image guaranteed to be indexed successfully by 
    index_<i,j>[:] +/- 1. This routine constructs an image with added border
    pixels... evilly, the index, 0 - 1, lands on the border because of Python's
    negative indexing convention.
    '''
    if np.issubdtype(image.dtype, float):
        image = image.astype(bool)
    image_i, image_j = np.argwhere(image.astype(bool)).transpose().astype(np.int32) + 1
    output_image = (np.ones(np.array(image.shape)+2,image.dtype) if border_value
                    else np.zeros(np.array(image.shape)+2, image.dtype))
    output_image[1:image.shape[0]+1, 1:image.shape[1]+1] = image
    return (image_i, image_j, output_image.astype(np.uint32))


def extract_from_image_lookup(orig_image, index_i, index_j):
    output = np.zeros(orig_image.shape, orig_image.dtype)
    output[index_i - 1, index_j - 1] = orig_image[index_i - 1, index_j - 1]
    return output

def ptrsize():
    '''The number of bytes in a pointer'''
    return sizeof(int *)

@cython.boundscheck(False)
def fill_labeled_holes_loop(
    np.ndarray[dtype = np.uint32_t, ndim = 1, negative_indices = False] i,
    np.ndarray[dtype = np.uint32_t, ndim = 1, negative_indices = False] j,
    np.ndarray[dtype = np.uint32_t, ndim = 1, negative_indices = False] idx,
    np.ndarray[dtype = np.uint32_t, ndim = 1, negative_indices = False] i_count,
    np.ndarray[dtype = np.uint8_t,  ndim = 1, negative_indices = False] is_not_hole,
    np.ndarray[dtype = np.uint32_t, ndim = 1, negative_indices = False] adjacent_non_hole,
    np.ndarray[dtype = np.uint32_t, ndim = 1, negative_indices = False] to_do,
    int lcount,
    int to_do_count):
    '''Run the graph loop portions of fill_labeled_holes
    
    i, j - the labels of pairs of touching objects
    idx - the index to the j for a particular i
    i_counts - the number of j for a particular i
    is_not_hole - on entry, a boolean array with one element per label. On exit,
                  True if the label is not a hole to be filled.
    adjacent_non_hole - on entry, an integer array set to zero with one
                        element per label. On exit, for each hole, the label
                        that should be used to fill it.
    to_do - one element per label. The initial non-hole labels (with to_do_count
            as the number on the list).
    lcount - all labels with a label # of lcount or below are objects, all above
             are background.
    '''
    cdef:
        int ii,jj,jidx
        int n = len(is_not_hole)
        int *p_idx = <int *>(idx.data)
        int *p_i_count = <int *>(i_count.data)
        int *p_to_do = <int *>(to_do.data)
        char *p_is_not_hole = <char *>(is_not_hole.data)
        int *p_adjacent_non_hole = <int *>(adjacent_non_hole.data)
        int *p_i = <int *>(i.data)
        int *p_j = <int *>(j.data)
    with nogil:
        while to_do_count > 0:
            ii = p_to_do[to_do_count - 1]
            to_do_count -= 1
            #
            # jj are the adjacencies to ii
            #
            for jidx from 0 <= jidx < p_i_count[ii]:
                jj = p_j[p_idx[ii] + jidx]
                if p_is_not_hole[jj] == 0:
                    if ii <= lcount:
                        #
                        # i labels an object. Label any object that is adjacent to
                        # a different object.
                        #
                        if p_adjacent_non_hole[jj] == 0:
                            p_adjacent_non_hole[jj] = ii
                            continue
                        elif p_adjacent_non_hole[jj] == ii:
                            continue
                    elif jj > lcount:
                        #
                        # i labels background. Label any foreground object touching it.
                        #
                        continue
                    p_is_not_hole[jj] = 1
                    p_to_do[to_do_count] = jj
                    to_do_count += 1
        #
        # Now we need to walk the graph of hole objects. There are holes
        # that are touching non-holes. These are the ones where adjacent_non_hole
        # is not zero. There are holes that are inside holes (objects inside holes)
        # and holes that are inside holes that are inside holes (holes inside
        # objects inside holes) and so on until you get to the bullseye of some
        # mega-archery-target.
        #
        to_do_count = 0
        for jj from 0 <= jj < n:
            if p_is_not_hole[jj] == 0 and p_adjacent_non_hole[jj] != 0:
                p_to_do[to_do_count] = jj
                to_do_count += 1
        while to_do_count > 0:
            ii = p_to_do[to_do_count - 1]
            to_do_count -= 1
            for jidx from 0 <= jidx < p_i_count[ii]:
                jj = p_j[p_idx[ii] + jidx]
                if p_is_not_hole[jj] == 0 and p_adjacent_non_hole[jj] == 0:
                    p_adjacent_non_hole[jj] = p_adjacent_non_hole[ii]
                    p_to_do[to_do_count] = jj
                    to_do_count += 1

@cython.boundscheck(False)
def trace_outlines(
    np.ndarray[dtype = np.uint32_t, ndim = 1, negative_indices = False] labels,
    np.ndarray[dtype = np.uint32_t, ndim = 1, negative_indices = False] firsts,
    np.ndarray[dtype = np.int32_t, ndim = 1, negative_indices = False] stride_table,
    np.ndarray[dtype = np.int32_t, ndim = 1, negative_indices = False] new_direction_table,
    np.ndarray[dtype = np.uint32_t, ndim = 1, negative_indices = False] output,
    np.ndarray[dtype = np.uint32_t, ndim = 1, negative_indices = False] output_count):
    '''Trace the border of contiguous, labeled objects
    
    Trace the border of contiguous, labeled objects using the maze traversal
    technique of conceptually putting your right hand on the wall and walking
    forward to visit each border pixel in turn. To prepare, the caller
    ravels a labels matrix of contiguous objects and supplies the index into
    the ravel of the point with the mininimum i**2 + j**2 (a northwest corner).
    The traversal order used when visiting points is supplied as a stride
    table that addresses, relatively, the pixel to be checked and visited.
    
    labels - a raveled labels matrix of the objects, zero-padded at the edges if
             necessary.
             
    firsts - the index of the northwest corner of each object to be checked.
    
    stride_table - gives the relative offset from the pixel of interest to
                   the pixel to be checked.
                   
    new_direction_table - gives the offset in the stride_table and 
                          new_direction_table of the next direction if
                          the entry is a hit.
                          
    output - an array of sufficient size to hold the indexes of the visited
             pixels in the outline. The worst case is that a pixel not in
             the interior can be visited twice.
             
    output_count - on output, the # of indexes per object in the tracing
    '''
    cdef:
        int n_labels = <int>(firsts.shape[0])
        int *p_labels = <int *>(labels.data)
        int *p_firsts = <int *>(firsts.data)
        int *p_stride_table = <int *>(stride_table.data)
        int *p_new_direction_table = <int *>(new_direction_table.data)
        int *p_output = <int *>(output.data)
        int *p_output_count = <int *>(output_count.data)
        int first_idx = 0
        int output_idx = 0
        int output_start
        int current_direction
        int current_location
        int current_label
        int test_location
        int output_end = <int>(output.shape[0])
        int overrun = 0
        int traversal_idx
        int traversal_entry

    assert len(stride_table) == 8
    assert len(new_direction_table) == 8
    if True:
        for first_idx from 0 <= first_idx < n_labels:
            #print "Label %d of %d" % (first_idx, n_labels)
            current_direction = 2
            current_location = p_firsts[first_idx]
            current_label = p_labels[current_location]
            output_start = output_idx
            #print "Label = %d, start = %d" % (current_label, current_location)
            while output_idx < output_end:
                p_output[output_idx] = current_location
                output_idx += 1
                for traversal_idx from 0 <= traversal_idx < 8:
                    traversal_entry = (traversal_idx + current_direction) & 7
                    test_location = current_location + p_stride_table[traversal_entry]
                    if p_labels[test_location] == current_label:
                        break
                else:
                    # Case: a single point can't find anything in the traversal
                    #print "No traversal"
                    break
                if test_location == p_firsts[first_idx]:
                    #print "Found %d" % p_firsts[first_idx]
                    break
                current_location = test_location
                current_direction = new_direction_table[traversal_entry]
                #print "Moving to %d, direction=%d" % (current_location, current_direction)
            else:
                overrun = 1
                break
            p_output_count[first_idx] = output_idx - output_start
    if overrun:
        raise IndexError("Output buffer overrun")
                
            



import numpy as np

def otsu(data, min_threshold=None, max_threshold=None,bins=256):
    """Compute a threshold using Otsu's method
    
    data           - an array of intensity values between zero and one
    min_threshold  - only consider thresholds above this minimum value
    max_threshold  - only consider thresholds below this maximum value
    bins           - we bin the data into this many equally-spaced bins, then pick
                     the bin index that optimizes the metric
    """
    assert min_threshold is None or max_threshold is None or min_threshold < max_threshold
    def constrain(threshold):
        if not min_threshold is None and threshold < min_threshold:
            threshold = min_threshold
        if not max_threshold is None and threshold > max_threshold:
            threshold = max_threshold
        return threshold
    
    data = np.atleast_1d(data)
    data = data[~ np.isnan(data)]
    if len(data) == 0:
        return (min_threshold if not min_threshold is None
                else max_threshold if not max_threshold is None
                else 0)
    elif len(data) == 1:
        return constrain(data[0])
    if bins > len(data):
        bins = len(data)
    data.sort()
    var = running_variance(data)
    rvar = np.flipud(running_variance(np.flipud(data))) 
    thresholds = data[1:len(data):len(data)/bins]
    score_low = (var[0:len(data)-1:len(data)/bins] * 
                 np.arange(0,len(data)-1,len(data)/bins))
    score_high = (rvar[1:len(data):len(data)/bins] *
                  (len(data) - np.arange(1,len(data),len(data)/bins)))
    scores = score_low + score_high
    if len(scores) == 0:
        return constrain(thresholds[0])
    index = np.argwhere(scores == scores.min()).flatten()
    if len(index)==0:
        return constrain(thresholds[0])
    #
    # Take the average of the thresholds to either side of
    # the chosen value to get an intermediate in cases where there is
    # a steep step between the background and foreground
    index = index[0]
    if index == 0:
        index_low = 0
    else:
        index_low = index-1
    if index == len(thresholds)-1:
        index_high = len(thresholds)-1
    else:
        index_high = index+1 
    return constrain((thresholds[index_low]+thresholds[index_high]) / 2)

def entropy(data, bins=256):
    """Compute a threshold using Ray's entropy measurement
    
    data           - an array of intensity values between zero and one
    bins           - we bin the data into this many equally-spaced bins, then pick
                     the bin index that optimizes the metric
    """
    
    data = np.atleast_1d(data)
    data = data[~ np.isnan(data)]
    if len(data) == 0:
        return 0
    elif len(data) == 1:
        return data[0]

    if bins > len(data):
        bins = len(data)
    data.sort()
    var = running_variance(data)+1.0/512.0
    rvar = np.flipud(running_variance(np.flipud(data)))+1.0/512.0 
    thresholds = data[1:len(data):len(data)/bins]
    w = np.arange(0,len(data)-1,len(data)/bins)
    score_low = w * np.log(var[0:len(data)-1:len(data)/bins] *
                           w * np.sqrt(2*np.pi*np.exp(1)))
    score_low[np.isnan(score_low)]=0
    
    w = len(data) - np.arange(1,len(data),len(data)/bins)
    score_high = w * np.log(rvar[1:len(data):len(data)/bins] * w *
                            np.sqrt(2*np.pi*np.exp(1)))
    score_high[np.isnan(score_high)]=0
    scores = score_low + score_high
    index = np.argwhere(scores == scores.min()).flatten()
    if len(index)==0:
        return thresholds[0]
    #
    # Take the average of the thresholds to either side of
    # the chosen value to get an intermediate in cases where there is
    # a steep step between the background and foreground
    index = index[0]
    if index == 0:
        index_low = 0
    else:
        index_low = index-1
    if index == len(thresholds)-1:
        index_high = len(thresholds)-1
    else:
        index_high = index+1 
    return (thresholds[index_low]+thresholds[index_high]) / 2

def otsu3(data, min_threshold=None, max_threshold=None,bins=128):    
    """Compute a threshold using a 3-category Otsu-like method
    
    data           - an array of intensity values between zero and one
    min_threshold  - only consider thresholds above this minimum value
    max_threshold  - only consider thresholds below this maximum value
    bins           - we bin the data into this many equally-spaced bins, then pick
                     the bin index that optimizes the metric
    
    We find the maximum weighted variance, breaking the histogram into
    three pieces.
    Returns the lower and upper thresholds
    """
    assert min_threshold is None or max_threshold is None or min_threshold < max_threshold
    
    #
    # Compute the running variance and reverse running variance.
    # 
    data = np.atleast_1d(data)
    data = data[~ np.isnan(data)]
    data.sort()
    if len(data) == 0:
        return 0
    var = running_variance(data)
    rvar = np.flipud(running_variance(np.flipud(data)))
    if bins > len(data):
        bins = len(data)
    bin_len = int(len(data)/bins) 
    thresholds = data[0:len(data):bin_len]
    score_low = (var[0:len(data):bin_len] * 
                 np.arange(0,len(data),bin_len))
    score_high = (rvar[0:len(data):bin_len] *
                  (len(data) - np.arange(0,len(data),bin_len)))
    #
    # Compute the middles
    #
    cs = data.cumsum()
    cs2 = (data**2).cumsum()
    i,j = np.mgrid[0:score_low.shape[0],0:score_high.shape[0]]*bin_len
    diff = (j-i).astype(float)
    w = diff
    mean = (cs[j] - cs[i]) / diff
    mean2 = (cs2[j] - cs2[i]) / diff
    score_middle = w * (mean2 - mean**2)
    score_middle[i >= j] = np.Inf
    score = score_low[i*bins/len(data)] + score_middle + score_high[j*bins/len(data)]
    best_score = np.min(score)
    best_i_j = np.argwhere(score==best_score)
    return (thresholds[best_i_j[0,0]],thresholds[best_i_j[0,1]])

def entropy3(data, bins=128):    
    """Compute a threshold using a 3-category Otsu-like method
    
    data           - an array of intensity values between zero and one
    bins           - we bin the data into this many equally-spaced bins, then pick
                     the bin index that optimizes the metric
    
    We find the maximum weighted variance, breaking the histogram into
    three pieces.
    Returns the lower and upper thresholds
    """
    #
    # Compute the running variance and reverse running variance.
    # 
    data = np.atleast_1d(data)
    data = data[~ np.isnan(data)]
    data.sort()
    if len(data) == 0:
        return 0
    var = running_variance(data)+1.0/512.0
    if bins > len(data):
        bins = len(data)
    bin_len = int(len(data)/bins) 
    thresholds = data[0:len(data):bin_len]
    score_low = entropy_score(var,bins)
    
    rvar = running_variance(np.flipud(data))+1.0/512.0 
    score_high = np.flipud(entropy_score(rvar,bins))
    #
    # Compute the middles
    #
    cs = data.cumsum()
    cs2 = (data**2).cumsum()
    i,j = np.mgrid[0:score_low.shape[0],0:score_high.shape[0]]*bin_len
    diff = (j-i).astype(float)
    w = diff / float(len(data))
    mean = (cs[j] - cs[i]) / diff
    mean2 = (cs2[j] - cs2[i]) / diff
    score_middle = entropy_score(mean2 - mean**2 + 1.0/512.0, bins, w, False)
    score_middle[(i >= j) | np.isnan(score_middle)] = np.Inf
    score = score_low[i/bin_len] + score_middle + score_high[j/bin_len]
    best_score = np.min(score)
    best_i_j = np.argwhere(score==best_score)
    return (thresholds[best_i_j[0,0]],thresholds[best_i_j[0,1]])

def entropy_score(var,bins, w=None, decimate=True):
    '''Compute entropy scores, given a variance and # of bins
    
    '''
    if w is None:
        n = len(var)
        w = np.arange(0,n,n/bins) / float(n)
    if decimate:
        n = len(var)
        var = var[0:n:n/bins]
    score = w * np.log(var * w * np.sqrt(2*np.pi*np.exp(1)))
    score[np.isnan(score)]=np.Inf
    return score
    

def weighted_variance(cs, cs2, lo, hi):
    if hi == lo:
        return np.Infinity
    w = (hi - lo) / float(len(cs))
    mean = (cs[hi] - cs[lo]) / (hi - lo)
    mean2 = (cs2[hi] - cs2[lo]) / (hi - lo)
    return w * (mean2 - mean**2)

def otsu_entropy(cs, cs2, lo, hi):
    if hi == lo:
        return np.Infinity
    w = (hi - lo) / float(len(cs))
    mean = (cs[hi] - cs[lo]) / (hi - lo)
    mean2 = (cs2[hi] - cs2[lo]) / (hi - lo)
    return w * (np.log (w * (mean2 - mean**2) * np.sqrt(2*np.pi*np.exp(1))))

def running_variance(x):
    '''Given a vector x, compute the variance for x[0:i]
    
    Thank you http://www.johndcook.com/standard_deviation.html
    S[i] = S[i-1]+(x[i]-mean[i-1])*(x[i]-mean[i])
    var(i) = S[i] / (i-1)
    '''
    n = len(x)
    # The mean of x[0:i]
    m = x.cumsum() / np.arange(1,n+1)
    # x[i]-mean[i-1] for i=1...
    x_minus_mprev = x[1:]-m[:-1]
    # x[i]-mean[i] for i=1...
    x_minus_m = x[1:]-m[1:]
    # s for i=1...
    s = (x_minus_mprev*x_minus_m).cumsum()
    var = s / np.arange(2,n+1)
    # Prepend Inf so we have a variance for x[0]
    return np.hstack(([0],var))



import numpy as np
cimport numpy as np
cimport cython

cdef extern from "Python.h":
    ctypedef int Py_intptr_t

cdef extern from "numpy/arrayobject.h":
    ctypedef class numpy.ndarray [object PyArrayObject]:
        cdef char *data
        cdef Py_intptr_t *strides
    cdef void import_array()
    cdef int  PyArray_ITEMSIZE(np.ndarray)

cdef extern from "stdlib.h":
    ctypedef unsigned long size_t
    void free(void *ptr)
    void *malloc(size_t size)

cdef extern from "string.h":
    void *memset(void *, int, int)

import_array()

##############################################################################
# 
# median_filter - implementation of constant-time median filter with
#                 octagonal shape. The algorithm is derived from
#                 Perreault, "Median Filtering in Constant Time",
#                 IEEE TRANSACTIONS ON IMAGE PROCESSING, VOL. 16, NO. 9,
#                 SEPTEMBER 2007.
#
# Inputs:
#    a 2d array of uint8's to be median filtered
#    a similarly shaped uint8 masking array with "1" indicating a significant
#         pixel and "0" indicating a pixel to be masked.
#
# Outputs:
#    a 2d median-filtered array
#
##############################################################################

DTYPE_UINT32 = np.uint32
DTYPE_BOOL = np.bool
ctypedef np.uint16_t pixel_count_t

###########
#
# Histograms
#
# There are five separate histograms for the octagonal filter and
# there are two levels (coarse = 16 values, fine = 256 values)
# per histogram. There are four histograms to maintain per position
# representing the four diagonals of the histogram plus one histogram
# for the straight side (which is used for adding and subtracting)
#
###########

cdef struct HistogramPiece:
    np.uint16_t coarse[16]
    np.uint16_t fine[256]

cdef struct Histogram:
    HistogramPiece top_left     # top-left corner
    HistogramPiece top_right    # top-right corner
    HistogramPiece edge         # leading/trailing edge
    HistogramPiece bottom_left  # bottom-left corner
    HistogramPiece bottom_right # bottom-right corner

# The pixel count has the number of pixels histogrammed in
# each of the five compartments for this position. This changes
# because of the mask
#
cdef struct PixelCount:
    pixel_count_t top_left
    pixel_count_t top_right
    pixel_count_t edge
    pixel_count_t bottom_left
    pixel_count_t bottom_right

#
# Stride + coordinates: the info we need when computing
# relative offsets from the octagon center
#
cdef struct SCoord:
    np.int32_t stride   # add the stride to the memory location
    np.int32_t x
    np.int32_t y

cdef struct Histograms:
    void *memory                # pointer to the allocated memory
    Histogram *histogram        # pointer to the histogram memory
    PixelCount *pixel_count     # pointer to the pixel count memory
    np.uint8_t *data            # pointer to the image data
    np.uint8_t *mask            # pointer to the image mask
    np.uint8_t *output          # pointer to the output array
    np.int32_t column_count     # number of columns represented by this structure
    np.int32_t stripe_length    # number of columns including "radius" before and after
    np.int32_t row_count        # number of rows available in image
    np.int32_t current_column   # the column being processed
    np.int32_t current_row      # the row being processed
    np.int32_t current_stride   # offset in data and mask to current location
    np.int32_t radius           # the "radius" of the octagon
    np.int32_t a_2              # 1/2 of the length of a side of the octagon
    # 
    #
    # The strides are the offsets in the array to the points that need to
    # be added or removed from a histogram to shift from the previous row
    # to the current row.
    # Listed here going clockwise from the trailing edge's top.
    # (-) = needs to be removed
    # (+) = needs to be added
    #
    #          -        -
    #         1.=========2
    #        1.           2
    #       +.             +-   Y
    #      |.               3   |
    #      |.               3   |
    #      -.               X  \|/
    #       5.             4    v
    #        5.           4
    #         +.=========+
    #
    #          x -->
    #
    SCoord last_top_left     # (-) left side of octagon's top - 1 row
    SCoord top_left          # (+) -1 row from trailing edge top
    SCoord last_top_right    # (-) right side of octagon's top - 1 col - 1 row
    SCoord top_right         # (+) -1 col -1 row from leading edge top
    SCoord last_leading_edge # (-) leading edge (right) top stride - 1 row
    SCoord leading_edge      # (+) leading edge bottom stride
    SCoord last_bottom_right # (-) leading edge bottom - 1 col
    SCoord bottom_right      # (+) right side of octagon's bottom - 1 col
    SCoord last_bottom_left  # (-) trailing edge bottom - 1 col
    SCoord bottom_left       # (+) left side of octagon's bottom - 1 col

    np.int32_t row_stride          # stride between one row and the next
    np.int32_t col_stride          # stride between one column and the next
    # The accumulator holds the running histogram
    #
    HistogramPiece accumulator
    #
    # The running count of pixels in the accumulator
    #
    np.uint32_t accumulator_count
    #
    # The percent of pixels within the octagon whose value is
    # less than or equal to the median-filtered value (e.g. for
    # median, this is 50, for lower quartile it's 25)
    #
    np.int32_t percent
    #
    # last_update_column keeps track of the column # of the last update
    # to the fine histogram accumulator. Short-term, the median
    # stays in one coarse block so only one fine histogram might
    # need to be updated
    #
    np.int32_t last_update_column[16]
    
############################################################################
#
# allocate_histograms - allocates the Histograms structure for the run
#
############################################################################
cdef Histograms *allocate_histograms(np.int32_t   rows, 
                                            np.int32_t   columns,
                                            np.int32_t   row_stride,
                                            np.int32_t   col_stride,
                                            np.int32_t   radius,
                                            np.int32_t   percent,
                                            np.uint8_t * data,
                                            np.uint8_t * mask,
                                            np.uint8_t * output):
    cdef:
        unsigned int adjusted_stripe_length = columns + 2*radius + 1
        unsigned int memory_size
        void *ptr
        Histograms *ph
        size_t roundoff
        int a
        SCoord *psc

    memory_size = (adjusted_stripe_length * 
                   (sizeof(Histogram) + sizeof(PixelCount))+
                   sizeof(Histograms)+32)
    ptr = malloc(memory_size)
    memset(ptr, 0, memory_size)
    ph  = <Histograms *>ptr
    if not ptr:
        return ph
    ph.memory = ptr
    ptr = <void *>(ph+1)
    ph.pixel_count = <PixelCount *>ptr
    ptr = <void *>(ph.pixel_count + adjusted_stripe_length)
    #
    # Align histogram memory to a 32-byte boundary
    #
    roundoff     = <size_t>ptr
    roundoff    += 31
    roundoff    -= roundoff % 32
    ptr          = <void *>roundoff
    ph.histogram = <Histogram *>ptr
    #
    # Fill in the statistical things we keep around
    #
    ph.column_count   = columns
    ph.row_count      = rows
    ph.current_column = -radius
    ph.stripe_length  = adjusted_stripe_length
    ph.current_row    = 0
    ph.radius         = radius
    ph.percent        = percent
    ph.row_stride     = row_stride
    ph.col_stride     = col_stride
    ph.data           = data
    ph.mask           = mask
    ph.output         = output
    #
    # Compute the coordinates of the significant points
    # (the SCoords)
    #
    # First, the length of a side of an octagon, compared
    # to what we call the radius is:
    #     2*r
    # ----------- =  a
    # (1+sqrt(2))
    #
    # a_2 is the offset from the center to each of the octagon
    # corners
    #
    a = <int>(<np.float64_t>radius * 2.0 / 2.414213)
    a_2 = a / 2
    if a_2 == 0:
        a_2 = 1
    ph.a_2 = a_2
    if radius <= a_2:
        radius = a_2+1
        ph.radius = radius

    ph.last_top_left.x = -a_2
    ph.last_top_left.y = -radius - 1

    ph.top_left.x = -radius
    ph.top_left.y = -a_2 - 1

    ph.last_top_right.x = a_2 - 1
    ph.last_top_right.y = -radius - 1

    ph.top_right.x = radius - 1
    ph.top_right.y = -a_2 - 1

    ph.last_leading_edge.x = radius
    ph.last_leading_edge.y = -a_2 - 1

    ph.leading_edge.x = radius
    ph.leading_edge.y = a_2

    ph.last_bottom_right.x = radius
    ph.last_bottom_right.y = a_2

    ph.bottom_right.x = a_2
    ph.bottom_right.y = radius

    ph.last_bottom_left.x = -radius-1
    ph.last_bottom_left.y = a_2

    ph.bottom_left.x = -a_2-1
    ph.bottom_left.y = radius

    #
    # Set the stride of each SCoord based on its x and y
    #
    set_stride(ph, &ph.last_top_left)
    set_stride(ph, &ph.top_left)
    set_stride(ph, &ph.last_top_right)
    set_stride(ph, &ph.top_right)
    set_stride(ph, &ph.last_leading_edge)
    set_stride(ph, &ph.leading_edge)
    set_stride(ph, &ph.last_bottom_left)
    set_stride(ph, &ph.bottom_left)
    set_stride(ph, &ph.last_bottom_right)
    set_stride(ph, &ph.bottom_right)

    return ph

############################################################################    
#
# free_histograms - frees the Histograms structure
#
############################################################################    
cdef void free_histograms(Histograms *ph):
    free(ph.memory)

############################################################################
#
# set_stride - set the stride of a SCoord from its X and Y
#
############################################################################

cdef void set_stride(Histograms *ph, SCoord *psc):
    psc.stride = psc.x * ph.col_stride + psc.y * ph.row_stride

############################################################################    
#
# <tl,tr,bl,br>_colidx - convert a column index into the histogram
#                        index for a diagonal
#
# The top-right and bottom left diagonals for one row at one column
# become the diagonals for the next column to the right for the next row.
# Conversely, the top-left and bottom right become the diagonals for the
# previous column.
#
# These functions use the current row number to find the index of
# a particular histogram taking this into account. The indices progress
# forward or backward as you go to successive rows.
#
# The histogram array is, in effect, a circular buffer, so the start
# offset is immaterial - we take advantage of this to make sure that
# the numbers computed before taking the modulus are all positive, including
# those that might be done for columns to the left of 0. We add 3* the radius
# here to account for a row of -radius, a column of -radius and a request for
# a column that is "radius" to the left.
#
############################################################################    
cdef inline np.int32_t tl_br_colidx(Histograms *ph, np.int32_t colidx):
    return (colidx + 3*ph.radius + ph.current_row)%ph.stripe_length

cdef inline np.int32_t tr_bl_colidx(Histograms *ph, np.int32_t colidx):
    return (colidx + 3*ph.radius + ph.row_count-ph.current_row) % ph.stripe_length

cdef inline np.int32_t leading_edge_colidx(Histograms *ph, np.int32_t colidx):
    return (colidx + 5*ph.radius) % ph.stripe_length

cdef inline np.int32_t trailing_edge_colidx(Histograms *ph, np.int32_t colidx):
    return (colidx + 3*ph.radius - 1) % ph.stripe_length
#
# add16 - add 16 consecutive integers
#
# Add an array of 16 16-bit integers to an accumulator of 16 16-bit integers
#
# TO_DO - optimize using SIMD instructions
#
cdef inline void add16(np.uint16_t *dest, np.uint16_t *src):
    cdef int i
    for i in range(16):
        dest[i] += src[i]

cdef inline void sub16(np.uint16_t *dest, np.uint16_t *src):
    cdef int i
    for i in range(16):
        dest[i] -= src[i]

############################################################################    
#
# accumulate_coarse_histogram - accumulate the coarse histogram
#                               at an index into the accumulator
#
# ph     - the Histograms structure that holds the accumulator
# colidx - the index of the column to add
#
############################################################################    
cdef inline void accumulate_coarse_histogram(Histograms *ph, np.int32_t colidx):
    cdef:
        int offset

    offset = tr_bl_colidx(ph, colidx)
    if ph.pixel_count[offset].top_right > 0:
        add16(ph.accumulator.coarse, ph.histogram[offset].top_right.coarse)
        ph.accumulator_count += ph.pixel_count[offset].top_right
    offset = leading_edge_colidx(ph, colidx)
    if ph.pixel_count[offset].edge > 0:
        add16(ph.accumulator.coarse, ph.histogram[offset].edge.coarse)
        ph.accumulator_count += ph.pixel_count[offset].edge
    offset = tl_br_colidx(ph, colidx)
    if ph.pixel_count[offset].bottom_right > 0:
        add16(ph.accumulator.coarse, ph.histogram[offset].bottom_right.coarse)
        ph.accumulator_count += ph.pixel_count[offset].bottom_right

############################################################################    
#
# deaccumulate_coarse_histogram - subtract the coarse histogram
#                                 for a given column
#
############################################################################    
cdef inline void deaccumulate_coarse_histogram(Histograms *ph, np.int32_t colidx):
    cdef:
        int offset
    #
    # The trailing diagonals don't appear until here
    #
    if colidx <= ph.a_2:
        return
    offset = tl_br_colidx(ph, colidx)
    if ph.pixel_count[offset].top_left > 0:
        sub16(ph.accumulator.coarse, ph.histogram[offset].top_left.coarse)
        ph.accumulator_count -= ph.pixel_count[offset].top_left
    #
    # The trailing edge doesn't appear from the border until here
    #
    if colidx > ph.radius:
        offset = trailing_edge_colidx(ph, colidx)
        if ph.pixel_count[offset].edge > 0:
            sub16(ph.accumulator.coarse, ph.histogram[offset].edge.coarse)
            ph.accumulator_count -= ph.pixel_count[offset].edge
    offset = tr_bl_colidx(ph, colidx)
    if ph.pixel_count[offset].bottom_left > 0:
        sub16(ph.accumulator.coarse, ph.histogram[offset].bottom_left.coarse)
        ph.accumulator_count -= ph.pixel_count[offset].bottom_left

############################################################################    
#
# accumulate_fine_histogram - accumulate one of the 16 fine histograms
#
############################################################################    
cdef inline void accumulate_fine_histogram(Histograms *ph, 
                                           np.int32_t colidx,
                                           np.uint32_t fineidx):
    cdef:
        int fineoffset = fineidx * 16
        int offset

    offset = tr_bl_colidx(ph, colidx)
    add16(ph.accumulator.fine+fineoffset, ph.histogram[offset].top_right.fine+fineoffset)
    offset = leading_edge_colidx(ph, colidx)
    add16(ph.accumulator.fine+fineoffset, ph.histogram[offset].edge.fine+fineoffset)
    offset = tl_br_colidx(ph, colidx)
    add16(ph.accumulator.fine+fineoffset, ph.histogram[offset].bottom_right.fine+fineoffset)

############################################################################    
#
# deaccumulate_fine_histogram - subtract one of the 16 fine histograms
#
############################################################################    
cdef inline void deaccumulate_fine_histogram(Histograms *ph, 
                                             np.int32_t colidx,
                                             np.uint32_t fineidx):
    cdef:
        int fineoffset = fineidx * 16
        int offset

    #
    # The trailing diagonals don't appear until here
    #
    if colidx < ph.a_2:
        return
    offset = tl_br_colidx(ph, colidx)
    sub16(ph.accumulator.fine+fineoffset, ph.histogram[offset].top_left.fine+fineoffset)
    if colidx >= ph.radius:
        offset = trailing_edge_colidx(ph, colidx)
        sub16(ph.accumulator.fine+fineoffset, ph.histogram[offset].edge.fine+fineoffset)
    offset = tr_bl_colidx(ph, colidx)
    sub16(ph.accumulator.fine+fineoffset, ph.histogram[offset].bottom_left.fine+fineoffset)
    
############################################################################    
#
# accumulate - add the leading edge and subtract the trailing edge
#
############################################################################    

cdef inline void accumulate(Histograms *ph):
    cdef:
        int i
        int j
        np.int32_t accumulator
    accumulate_coarse_histogram(ph, ph.current_column)
    deaccumulate_coarse_histogram(ph, ph.current_column)

############################################################################    
#
# update_fine - update one of the fine histograms to the current column
#
# The code has two choices:
#    redo the fine histogram from scratch - this involves accumulating
#         the entire histogram from the top_left.x to the top_right.x,
#         the center (edge) histogram from the trailing edge x to the
#         top_left.x and then computing a histogram of all points between
#         the trailing edge top, the point, (top_left.x,trailing edge top.y)
#         and the top_right and the corresponding triangle in the octagon's
#         lower half.
#
#    accumulate and deaccumulate within the fine histogram from the last
#    column computed.
#
#    The code below only implements the accumulate; redo and the code
#    to choose remains to be done.
############################################################################    

cdef inline void update_fine(Histograms *ph, int fineidx):
    cdef:
        int first_update_column = ph.last_update_column[fineidx]+1
        int update_limit        = ph.current_column+1
        int i
 
    for i in range(first_update_column, update_limit):
        accumulate_fine_histogram(ph, i, fineidx)
        deaccumulate_fine_histogram(ph, i, fineidx)
    ph.last_update_column[fineidx] = ph.current_column

############################################################################    
#
# update_histogram - update the coarse and fine levels of a histogram
#                    based on addition of one value and subtraction of another
#
# ph         - Histograms pointer (for access to row_count, column_count)
# hist_piece - coarse and fine histogram to update
# pixel_count- pointer to pixel counter for histogram
# last_coord - coordinate and stride of pixel to remove
# coord      - coordinate and stride of pixel to add
# 
############################################################################    
cdef inline void update_histogram(Histograms *ph,
                                  HistogramPiece *hist_piece,
                                  pixel_count_t *pixel_count,
                                  SCoord *last_coord,
                                  SCoord *coord):
    cdef:
        np.int32_t current_column = ph.current_column
        np.int32_t current_row    = ph.current_row
        np.int32_t current_stride = ph.current_stride
        np.int32_t column_count   = ph.column_count
        np.int32_t row_count      = ph.row_count
        np.uint8_t value
        np.int32_t stride
        np.int32_t x
        np.int32_t y

    x      = last_coord.x + current_column
    y      = last_coord.y + current_row
    stride = current_stride+last_coord.stride

    if (x >= 0 and x < column_count and
        y >= 0 and y < row_count and
        ph.mask[stride]):
        value = ph.data[stride]
        pixel_count[0] -= 1
        hist_piece.fine[value] -= 1
        hist_piece.coarse[value / 16] -= 1

    x      = coord.x + current_column
    y      = coord.y + current_row
    stride = current_stride + coord.stride

    if (x >= 0 and x < column_count and
        y >= 0 and y < row_count and
        ph.mask[stride]):
        value = ph.data[stride]
        pixel_count[0] += 1
        hist_piece.fine[value] += 1
        hist_piece.coarse[value / 16] += 1

############################################################################    
#
# update_current_location - update the histograms at the current location
#
############################################################################    
cdef inline void update_current_location(Histograms *ph):
    cdef:
        np.int32_t current_column   = ph.current_column
        np.int32_t radius           = ph.radius
        np.int32_t top_left_off     = tl_br_colidx(ph, current_column)
        np.int32_t top_right_off    = tr_bl_colidx(ph, current_column)
        np.int32_t bottom_left_off  = tr_bl_colidx(ph, current_column)
        np.int32_t bottom_right_off = tl_br_colidx(ph, current_column)
        np.int32_t leading_edge_off = leading_edge_colidx(ph, current_column)
        np.int32_t *coarse_histogram
        np.int32_t *fine_histogram
        np.int32_t last_xoff
        np.int32_t last_yoff
        np.int32_t last_stride
        np.int32_t xoff
        np.int32_t yoff
        np.int32_t stride

    update_histogram(ph, &ph.histogram[top_left_off].top_left,
                     &ph.pixel_count[top_left_off].top_left,
                     &ph.last_top_left,
                     &ph.top_left)

    update_histogram(ph, &ph.histogram[top_right_off].top_right,
                     &ph.pixel_count[top_right_off].top_right,
                     &ph.last_top_right,
                     &ph.top_right)

    update_histogram(ph, &ph.histogram[bottom_left_off].bottom_left,
                     &ph.pixel_count[bottom_left_off].bottom_left,
                     &ph.last_bottom_left,
                     &ph.bottom_left)

    update_histogram(ph, &ph.histogram[bottom_right_off].bottom_right,
                     &ph.pixel_count[bottom_right_off].bottom_right,
                     &ph.last_bottom_right,
                     &ph.bottom_right)

    update_histogram(ph, &ph.histogram[leading_edge_off].edge,
                     &ph.pixel_count[leading_edge_off].edge,
                     &ph.last_leading_edge,
                     &ph.leading_edge)

############################################################################
#
# find_median - search the current accumulator for the median
#
############################################################################

cdef inline np.uint8_t find_median(Histograms *ph):
    cdef:
        np.uint32_t pixels_below      # of pixels below the median
        int i
        int j
        int k
        np.uint32_t accumulator

    if ph.accumulator_count == 0:
        return 0
    pixels_below = (ph.accumulator_count * ph.percent + 50) / 100 # +50 for roundoff
    if pixels_below > 0:
        pixels_below -= 1
    accumulator = 0
    for i in range(16):
        accumulator += ph.accumulator.coarse[i]
        if accumulator > pixels_below:
            break
    accumulator -= ph.accumulator.coarse[i]
    update_fine(ph, i)
    for j in range(i*16,(i+1)*16):
        accumulator += ph.accumulator.fine[j]
        if accumulator > pixels_below:
            return <np.uint8_t> j
    return 0

############################################################################
#
# c_median_filter - median filter algorithm
#
# rows    - # of rows in each array
# columns - # of columns in each array
# row_stride - stride from one row to the next in each array
# col_stride - stride from one column to the next in each array
# radius - radius of circle inscribed into octagon
# percent - "median" cutoff: 50 = median, 25 = lower quartile, etc
# data - array of image pixels to be filtered
# mask - mask of significant pixels
# output - array to be filled with filtered pixels
#
############################################################################
cdef int c_median_filter(np.int32_t   rows, 
                         np.int32_t   columns,
                         np.int32_t   row_stride,
                         np.int32_t   col_stride,
                         np.int32_t   radius,
                         np.int32_t   percent,
                         np.uint8_t * data,
                         np.uint8_t * mask,
                         np.uint8_t * output):
    cdef:
        Histograms *ph
        Histogram  *phistogram
        int row
        int col
        int i
        np.int32_t top_left_off
        np.int32_t top_right_off
        np.int32_t bottom_left_off
        np.int32_t bottom_right_off

    ph = allocate_histograms(rows, columns, row_stride, col_stride,
                             radius, percent, data, mask, output)
    if not ph:
       return 1

    for row in range(-radius, rows):
        #
        # Initialize the starting diagonal histograms to zero. The leading
        # and trailing histograms descend from above and so are initialized
        # when memory is initially set to zero. The diagonals move in
        # from the left (top left and bottom right) and right (top right
        # and bottom left). One of each needs to be initialized at the
        # start of each row
        #
        tl_br_off     = tl_br_colidx(ph, -radius)
        tr_bl_off     = tr_bl_colidx(ph, columns+radius-1)

        memset(&ph.histogram[tl_br_off].top_left, 0, sizeof(HistogramPiece))
        memset(&ph.histogram[tl_br_off].bottom_right, 0, sizeof(HistogramPiece))
        memset(&ph.histogram[tr_bl_off].top_right, 0, sizeof(HistogramPiece))
        memset(&ph.histogram[tr_bl_off].bottom_left, 0, sizeof(HistogramPiece))
        ph.pixel_count[tl_br_off].top_left     = 0
        ph.pixel_count[tl_br_off].bottom_right = 0
        ph.pixel_count[tr_bl_off].top_right    = 0
        ph.pixel_count[tr_bl_off].bottom_left  = 0
        #
        # Initialize the accumulator (octagon histogram) to zero
        #
        memset(&ph.accumulator, 0, sizeof(ph.accumulator))
        ph.accumulator_count = 0
        for i in range(16):
            ph.last_update_column[i] = -radius-1
        #
        # Initialize the current stride to the beginning of the row
        #
        ph.current_row = row
        #
        # Update locations and coarse accumulator for the octagon
        # for points before 0
        #
        for col in range(-radius, 0 if row >=0 else columns+radius):
            ph.current_column = col
            ph.current_stride = row * row_stride + col * col_stride
            update_current_location(ph)
            accumulate(ph)
        #
        # Update locations and coarse accumulator and compute
        # the median for points between 0 and "columns"
        #
        if row >= 0:
            for col in range(0, columns):
                ph.current_column = col
                ph.current_stride = row * row_stride + col * col_stride
                update_current_location(ph)
                accumulate(ph)
                ph.output[ph.current_stride] = find_median(ph)
            for col in range(columns, columns+radius):
                ph.current_column = col
                ph.current_stride = row * row_stride + col * col_stride
                update_current_location(ph)

    
    free_histograms(ph)
    return 0

def median_filter(np.ndarray[dtype=np.uint8_t, ndim=2, negative_indices=False, mode='c'] data,
                  np.ndarray[dtype=np.uint8_t, ndim=2, negative_indices=False, mode='c'] mask,
                  np.ndarray[dtype=np.uint8_t, ndim=2, negative_indices=False, mode='c'] output,
                  int radius,
                  np.int32_t percent):
    """Median filter with octagon shape and masking

    data - a 2d array containing the image data
    mask - a 2d array of 1=significant pixel, 0=masked
           similarly shaped to "data"
    output - a 2d array that will hold the output of this operation
             similarly shaped to "data"
    radius - the radius of the inscribed circle to the octagon
    percent - sort the unmasked pixels within the octagon into
              an array (conceptually) and take the value indexed
              by the size of that array times the percent divided by 100.
              50 gives the median
    """
    if percent < 0:
        raise ValueError('Median filter percent = %d is less than zero'%percent)
    if percent > 100:
        raise ValueError('Median filter percent = %d is greater than 100'%percent)
    if data.shape[0] != mask.shape[0] or data.shape[1] != mask.shape[1]:
        raise ValueError('Data shape (%d,%d) is not mask shape (%d,%d)'%
                         (data.shape[0],data.shape[1],
                          mask.shape[0], mask.shape[1]))
    if data.shape[0] != output.shape[0] or data.shape[1] != output.shape[1]:
        raise ValueError('Data shape (%d,%d) is not output shape (%d,%d)'%
                         (data.shape[0],data.shape[1],
                          output.shape[0], output.shape[1]))
    if c_median_filter(data.shape[0],   data.shape[1],
                       data.strides[0], data.strides[1],
                       radius, percent,
                       <np.uint8_t *>data.data,
                       <np.uint8_t *>mask.data, 
                       <np.uint8_t *>output.data):
        raise MemoryError('Failed to allocate scratchpad memory')

@cython.boundscheck(False)
def masked_convolution(np.ndarray[dtype=np.float64_t, ndim=2, negative_indices=False, mode='c'] data,
                       np.ndarray[dtype=np.uint8_t, ndim=2, negative_indices=False, mode='c'] mask,
                       np.ndarray[dtype=np.float64_t, ndim=2, negative_indices=False, mode='c'] kernel):
    """Convolution respecting a mask

    data - a 2d array containing the image data
    mask - a mask of relevant points.
    kernel - a square convolution kernel of odd dimension
    """
    cdef:
        np.float64_t *pkernel
        np.float64_t *pimage
        np.uint8_t   *pmask
        np.float64_t *poutput
        int          kernel_stride
        int          image_stride
        int          mask_stride
        int          mask_offset
        int          kernel_width
        int          kernel_half_width
        int          i,j,ik,jk
        int          istride,mstride,kstride
        np.float64_t accumulator
        np.ndarray[dtype=np.uint8_t, ndim=2, negative_indices=False, mode='c'] big_mask
        np.ndarray[dtype=np.float64_t, ndim=2, negative_indices=False, mode='c'] output

    assert kernel.shape[0] % 2 == 1, "Kernel shape must be odd"
    assert kernel.shape[0]==kernel.shape[1], "Kernel must be square"
    assert mask.shape[0]==data.shape[0]
    assert mask.shape[1]==data.shape[1]
    kernel_width = kernel.shape[0]
    kernel_half_width = kernel_width / 2
    big_mask = np.zeros((data.shape[0]+kernel_width, data.shape[1]+kernel_width), np.uint8)
    output   = np.zeros((data.shape[0],data.shape[1]), data.dtype)
    big_mask[kernel_half_width:kernel_half_width+data.shape[0],
             kernel_half_width:kernel_half_width+data.shape[1]] = mask
    #
    # stride in number of elements across the i direction
    #
    istride = data.strides[0] / PyArray_ITEMSIZE(data)
    mstride = big_mask.strides[0] / PyArray_ITEMSIZE(big_mask)
    kstride = kernel.strides[0] / PyArray_ITEMSIZE(kernel)
    #
    # pointers to data. pmask is offset to point at the 0,0 element
    # pkernel is offset to point at the middle of the kernel
    #
    pmask   = <np.uint8_t *>(big_mask.data + kernel_half_width *
                             (big_mask.strides[0] + big_mask.strides[1]))
    pimage  = <np.float64_t *>(data.data)
    pkernel = <np.float64_t *>(kernel.data)+(kstride+1) * kernel_half_width
    poutput = <np.float64_t *>(output.data)
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            if pmask[i*mstride+j] == 0:
                continue
            accumulator = 0
            for ik in range(-kernel_half_width,kernel_half_width+1):
                for jk in range(-kernel_half_width,kernel_half_width+1):
                    if pmask[(i+ik)*mstride+j+jk] != 0:
                        accumulator += (pkernel[ik*kstride+jk] *
                                        pimage[(i+ik)*istride+j+jk])
            poutput[i*istride+j] = accumulator
    return output

@cython.boundscheck(False)
@cython.cdivision(True)
def paeth_decoder(
    np.ndarray[dtype=np.uint8_t, ndim=3, negative_indices=False, mode='c'] x,
    np.int32_t raster_count):
    '''Paeth decoder - reverse Paeth filter
    
    x: matrix of bytes. The first dimension indexes rasters. The second
       dimension indexes pixel positions (stride = 24 for interleaved color,
       = 8 for monochrome). The third dimension indexes the bytes within
       the pixel. If your image consists of multiple planes, you should
       combine the planar dimensions on input by changing the image shape.
       
    raster_count: # of rasters in a plane
    
    Given a 2-dimensional array of unsigned bytes, the Paeth filter
    looks at 4 elements
    
    C B
    A x
    
    p = A+B-C
    estimate = A if abs(p-A) <= abs(p-B) and similarly for C
             = B if abs(p-B) <= abs(p-C)
             = C otherwise
    x += estimate if reverse
    
    Citation: http://www.w3.org/TR/PNG-Filters.html
    '''
    cdef:
        np.int32_t raster_stride = x.strides[0]
        np.int32_t pixel_stride = x.strides[1]
        unsigned char *ptr = <unsigned char *>x.data
        np.int32_t a,b,c,p,pa,pb,pc,estimate
        np.int32_t i,j,k
        np.int32_t imax = x.shape[0]
        np.int32_t jmax = x.shape[1]
        np.int32_t kmax = x.shape[2]
        np.int32_t raster_number
        np.int32_t plane_number
        
    with nogil:
        for i from 0<=i<imax:
            raster_number = i % raster_count
            for j from 0<=j<jmax:
                for k from 0<=k<kmax:
                    if raster_number == 0:
                        b = c = 0
                        if j==0:
                            a = 0
                        else:
                            a = ptr[-pixel_stride]
                    else:
                        b = ptr[-raster_stride]
                        if j==0:
                           a = c = 0
                        else:
                           a = ptr[-pixel_stride]
                           c = ptr[-raster_stride-pixel_stride]
                    p = a + b - c
                    pa = (a-p) if (a>p) else (p-a)
                    pb = (b-p) if (b>p) else (p-b)
                    pc = (c-p) if (c>p) else (p-c)
                    if (pa <= pb) and (pa <= pc):
                        estimate = a
                    elif (pb <= pc):
                        estimate = b
                    else:
                        estimate = c
                    ptr[0] += estimate
                    ptr += 1



import cython
import numpy as np
cimport numpy as np
ctypedef np.int32_t DTYPE_t  # 32-bit pixel positions and labels


cdef enum:
    DEBUG = 0

# Does the path a->b->c form a convexity in the plane?
cdef inline int CONVEX(int a_i, int a_j,
                       int b_i, int b_j,
                       int c_i, int c_j) nogil:
    cdef int ab_i, ab_j, bc_i, bc_j, cross
    ab_i = b_i - a_i
    ab_j = b_j - a_j
    bc_i = c_i - b_i
    bc_j = c_j - b_j
    # note that x is j, y is i
    cross = (ab_j * bc_i - bc_j * ab_i)
    if cross > 0:
        return 1
    if cross < 0:
        return 0
    # special case for a U-turn at the end of the points
    return ((b_j > a_j) and (b_j > c_j))

@cython.boundscheck(False)
@cython.wraparound(False)
def convex_hull_ijv(in_labels_ijv,
                    indexes_in):
    # reorder by v, then j, then i.  Note: we will overwrite this array with
    # the output.  The sorting allocates a new copy, and it's guaranteed to be
    # large enough for the convex hulls.
    cdef np.ndarray[DTYPE_t, ndim=2] labels_ijv = in_labels_ijv.astype(np.int32)[np.lexsort(in_labels_ijv.T), :]
    assert np.all(labels_ijv >= 0), "All ijv values must be >=0"
    cdef np.ndarray[DTYPE_t, ndim=1] indexes = np.asarray(indexes_in, np.int32).ravel()
    assert np.all(indexes >= 0), "All indexes must be >= 0"
    # declaration of local variables
    cdef int num_indexes, max_i, max_j, max_label, pixidx, outidx, cur_req, cur_label
    cdef int start_j, cur_pix_i, cur_pix_j, end_j, start_idx, num_vertices
    cdef int need_last_upper_point
    cdef int num_emitted
    # an indirect sorting array for indexes
    cdef np.ndarray[DTYPE_t, ndim=1] indexes_reorder = np.argsort(indexes).astype(np.int32)
    num_indexes = len(indexes)
    # find the maximums
    max_i, max_j, max_label = labels_ijv.max(axis=0)
    # allocate the upper and lower vertex buffers, and initialize them to extreme values
    cdef np.ndarray[DTYPE_t, ndim=1] upper = np.empty(max_j + 1, np.int32)
    cdef np.ndarray[DTYPE_t, ndim=1] lower = np.empty(max_j + 1, np.int32)
    cdef np.ndarray[DTYPE_t, ndim=1] vertex_counts = np.zeros(num_indexes, np.int32)
    cdef np.ndarray[DTYPE_t, ndim=1] hull_offsets = np.zeros(num_indexes, np.int32)
    pixidx = 0  # the next input pixel we'll process
    outidx = 0  # the next row to be written in the output
    for cur_req in range(num_indexes):
        cur_label = indexes[indexes_reorder[cur_req]]
        while (cur_label <= max_label) and (labels_ijv[pixidx, 2] < cur_label):
            pixidx += 1
            if pixidx == labels_ijv.shape[0]:
                break
        hull_offsets[cur_req] = outidx
        if (pixidx == labels_ijv.shape[0]) or (cur_label != labels_ijv[pixidx, 2]):
            # cur_label's hull will have 0 points
            continue
        start_idx = pixidx
        while pixidx < labels_ijv.shape[0] and cur_label == labels_ijv[pixidx, 2]:
            pixidx += 1
        num_vertices = pixidx - start_idx
        start_j = labels_ijv[start_idx, 1]
        end_j = labels_ijv[pixidx - 1, 1]
        upper[start_j:(end_j+1)] = -1
        lower[start_j:(end_j+1)] = max_i + 1
        for start_idx <= pixidx < (start_idx + num_vertices):
            cur_pix_i = labels_ijv[pixidx, 0]
            cur_pix_j = labels_ijv[pixidx, 1]
            if upper[cur_pix_j] < cur_pix_i:
                upper[cur_pix_j] = cur_pix_i
            if lower[cur_pix_j] > cur_pix_i:
                lower[cur_pix_j] = cur_pix_i
        pixidx = start_idx + num_vertices
        if DEBUG:
            print "STARTJ/END_J", start_j, end_j
            print "LOWER", lower[start_j:end_j + 1]
            print "UPPER", upper[start_j:end_j + 1]

        # At this point, the upper and lower buffers have the extreme high/low
        # points, so we just need to convexify them.  We have copied them out
        # of the labels_ijv array, so we write out the hull into that array
        # (via its alias "out").  We are careful about memory when we do so, to
        # make sure we don't invalidate the next entry in labels_ijv.
        #
        # We assume that the j-coordinates are dense.  If this assumption is
        # violated, we should re-walk the pixidx values we just copied to only
        # deal with columns that actually have points.
        #
        # Produce hull in counter-clockwise order, starting with lower
        # envelope.  Reset the envelopes as we do so.

        # Macro-type functions from Python version
        # def CONVEX(pt_i, pt_j, nemitted):
        #     # We're walking CCW, so left turns are convex
        #     d_i_prev, d_j_prev, z = out[out_base_idx + nemitted - 1, :] - out[out_base_idx + nemitted - 2, :]
        #     d_i_cur = pt_i - out[out_base_idx + nemitted - 1, 0]
        #     d_j_cur = pt_j - out[out_base_idx + nemitted - 1, 1]
        #     # note that x is j, y is i
        #     return (d_j_prev * d_i_cur - d_j_cur * d_i_prev) > 0
        #
        # def EMIT(pt_i, pt_j, nemitted):
        #     while (nemitted >= 2) and not CONVEX(pt_i, pt_j, nemitted):
        #         # The point we emitted just before this one created a
        #         # concavity (or is co-linear).  Prune it.
        #         #XXX print "BACKUP"
        #         nemitted -= 1
        #     # The point is convex or we haven't emitted enough points to check.
        #     #XXX print "writing point", nemitted, pt_i, pt_j
        #     out[out_base_idx + nemitted, :] = (pt_i, pt_j, cur_label)
        #     return nemitted + 1

        need_last_upper_point = (lower[start_j] != upper[start_j])
        num_emitted = 0
        for envelope_j in range(start_j, end_j + 1):
            if lower[envelope_j] < max_i + 1:
                # MACRO EMIT(lower[envelope_j], envelope_j, num_emitted)
                while (num_emitted >= 2) and not CONVEX(labels_ijv[outidx + num_emitted - 2, 0], labels_ijv[outidx + num_emitted - 2, 1],
                                                        labels_ijv[outidx + num_emitted - 1, 0], labels_ijv[outidx + num_emitted - 1, 1],
                                                        lower[envelope_j], envelope_j):
                    # The point we emitted just before this one created a concavity (or is co-linear).  Prune it.
                    if DEBUG:
                        print "PRUNE"
                    num_emitted -= 1
                labels_ijv[outidx + num_emitted, 0] = lower[envelope_j]
                labels_ijv[outidx + num_emitted, 1] = envelope_j
                if DEBUG:
                    print "ADD", (lower[envelope_j], envelope_j, cur_label)
                num_emitted += 1
                # END MACRO
                lower[envelope_j] = max_i + 1
        for envelope_j in range(end_j, start_j, -1):
            if upper[envelope_j] > -1:
                # MACRO EMIT(upper[envelope_j], envelope_j, num_emitted)
                while (num_emitted >= 2) and not CONVEX(labels_ijv[outidx + num_emitted - 2, 0], labels_ijv[outidx + num_emitted - 2, 1],
                                                        labels_ijv[outidx + num_emitted - 1, 0], labels_ijv[outidx + num_emitted - 1, 1],
                                                        upper[envelope_j], envelope_j):
                    # The point we emitted just before this one created a concavity (or is co-linear).  Prune it.
                    if DEBUG:
                        print "PRUNE"
                    num_emitted -= 1
                if pixidx > outidx+num_emitted:
                    # The last point can get temporarily pushed on the list
                    # twice only to be taken off below. So this is more than
                    # just keeping a segfault from happening.
                    labels_ijv[outidx + num_emitted, 0] = upper[envelope_j]
                    labels_ijv[outidx + num_emitted, 1] = envelope_j
                    if DEBUG:
                        print "ADD", (upper[envelope_j], envelope_j, cur_label)
                    num_emitted += 1
                # END MACRO
                upper[envelope_j] = -1
        # Even if we don't add the start point, we still might need to prune.
        # MACRO EMIT(upper[start_j], envelope_j, num_emitted)
        while (num_emitted >= 2) and not CONVEX(labels_ijv[outidx + num_emitted - 2, 0], labels_ijv[outidx + num_emitted - 2, 1],
                                                labels_ijv[outidx + num_emitted - 1, 0], labels_ijv[outidx + num_emitted - 1, 1],
                                                upper[start_j], start_j):
            # The point we emitted just before this one created a concavity (or is co-linear).  Prune it.
            if DEBUG:
                print "PRUNE"
            num_emitted -= 1
        if need_last_upper_point:
            labels_ijv[outidx + num_emitted, 0] = upper[start_j]
            labels_ijv[outidx + num_emitted, 1] = start_j
            if DEBUG:
                print "ADD", (upper[start_j], start_j, cur_label)
            num_emitted += 1
            # END MACRO
        upper[start_j] = -1
        # advance the output index
        vertex_counts[cur_req] = num_emitted
        outidx += num_emitted
    # sort by the requested order
    cdef np.ndarray[DTYPE_t, ndim=2] reordered = np.empty((outidx, 3), np.int32)
    cdef np.ndarray[DTYPE_t, ndim=1] reordered_counts = np.empty(num_indexes, np.int32)
    cdef np.ndarray[DTYPE_t, ndim=1] indexes_unreorder = np.argsort(indexes_reorder).astype(np.int32)
    cdef int reordered_idx, reordered_num, count, src_start, dest_start, tmpidx
    reordered_idx = 0
    if DEBUG:
        print "indexes", indexes
        print "reorder", indexes_reorder
        print "unreorder", indexes_unreorder
        print "Hull offsets", hull_offsets
    # Output in the order that the indices were passed in
    for reordered_num in range(num_indexes):
        cur_label = indexes[reordered_num]
        count = vertex_counts[indexes_unreorder[reordered_num]]
        src_start = hull_offsets[indexes_unreorder[reordered_num]]
        if DEBUG:
            print "writing index", cur_label, count, src_start
        dest_start = reordered_idx
        for tmpidx in range(count):
            # Reorder columns to match what CellProfiler expects.
            reordered[dest_start + tmpidx, 0] = cur_label
            reordered[dest_start + tmpidx, 1] = labels_ijv[src_start + tmpidx, 0]
            reordered[dest_start + tmpidx, 2] = labels_ijv[src_start + tmpidx, 1]
        reordered_idx += count
        reordered_counts[reordered_num] = count
    return reordered, reordered_counts



import Cython.Build
import pkg_resources
import setuptools
import setuptools.command.build_ext
import setuptools.command.test
import sys


class BuildExtension(setuptools.command.build_ext.build_ext):
    def build_extensions(self):
        numpy_includes = pkg_resources.resource_filename("numpy", "core/include")

        for extension in self.extensions:
            if hasattr(extension, "include_dirs") and numpy_includes not in extension.include_dirs:
                extension.include_dirs.append(numpy_includes)

            extension.include_dirs.append("centrosome/include")

        setuptools.command.build_ext.build_ext.build_extensions(self)


class Test(setuptools.command.test.test):
    user_options = [
        ("pytest-args=", "a", "Arguments to pass to py.test")
    ]

    def initialize_options(self):
        setuptools.command.test.test.initialize_options(self)

        self.pytest_args = []

    def finalize_options(self):
        setuptools.command.test.test.finalize_options(self)

        self.test_args = []

        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)

        sys.exit(errno)


setuptools.setup(
    author="Allen Goodman",
    author_email="agoodman@broadinstitute.org",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: C",
        "Programming Language :: C++",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 2",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering",
    ],
    cmdclass={
        "build_ext": BuildExtension,
        "test": Test
    },
    description="An open source image processing library",
    ext_modules=Cython.Build.cythonize([
        setuptools.Extension(
            name="_cpmorphology",
            sources=[
                "centrosome/src/_cpmorphology.c",
            ]
        ),
        setuptools.Extension(
            name="_propagate",
            sources=[
                "centrosome/_propagate.pyx",
            ],
        ),
        setuptools.Extension(
            name="*",
            language="c++",
            sources=[
                "centrosome/*.pyx",
            ],
        ),
    ]),
    install_requires=[
        "matplotlib",
        "numpy",
        "pillow",
        "scikit-image",
        "scipy",
    ],
    keywords="",
    license="BSD",
    long_description="",
    name="centrosome",
    packages=[
        "centrosome"
    ],
    setup_requires=[
        "cython",
        "numpy",
        "pytest",
    ],
    tests_require=[
        "pytest",
    ],
    url="https://github.com/CellProfiler/centrosome",
    version="1.0.8",
)



import numpy as np

from _lapjv import reduction_transfer
from _lapjv import augmenting_row_reduction
from _lapjv import augment

def lapjv(i, j, costs, wants_dual_variables = False, augmenting_row_reductions = 2):
    '''Sparse linear assignment solution using Jonker-Volgenant algorithm
    
    i,j - similarly-sized vectors that pair the object at index i[n] with
          the object at index j[j]
          
    costs - a vector of similar size to i and j that is the cost of pairing
            i[n] with j[n].
            
    wants_dual_variables - the dual problem reduces the costs using two
            vectors, u[i] and v[j] where the solution is the maximum value of
            np.sum(u) + np.sum(v) where cost[i,j] - u[i] - v[j] >=  0.
            Set wants_dual_variables to True to have u and v returned in
            addition to the assignments.
            
    augmenting_row_reductions - the authors suggest that augmenting row reduction
            be performed twice to optimize the u and v before the augmenting
            stage. The caller can choose a different number of reductions
            by supplying a different value here.

    All costs not appearing in i,j are taken as infinite. Each i in the range,
    0 to max(i) must appear at least once and similarly for j.
    
    returns (x, y), the pairs of assignments that represent the solution
    or (x, y, u, v) if the dual variables are requested.
    '''
    import os
    i = np.atleast_1d(i).astype(int)
    j = np.atleast_1d(j).astype(int)
    costs = np.atleast_1d(costs)
    
    assert len(i) == len(j), "i and j must be the same length"
    assert len(i) == len(costs), "costs must be the same length as i"

    #
    # Find the number of i with non-infinite cost for each j
    #
    j_count = np.bincount(j)
    assert not np.any(j_count == 0), "all j must be paired with at least one i"
    #
    # if you order anything by j, this is an index to the minimum for each j
    #
    j_index = np.hstack([[0], np.cumsum(j_count[:-1])])
    #
    # Likewise for i
    #
    i_count = np.bincount(i)
    assert not np.any(i_count == 0), "all i must be paired with at least one j"
    i_index = np.hstack([[0], np.cumsum(i_count[:-1])])
    
    n = len(j_count) # dimension of the square cost matrix
    assert n == len(i_count), "There must be the same number of unique i and j"
    
    # # # # # # # #
    #
    # Variable initialization:
    #
    # The output variables:
    #
    # x - for each i, the assigned j. -1 indicates uninitialized
    # y - for each j, the assigned i
    # u, v - the dual variables
    #
    # A value of x = n or y = n means "unassigned"
    #
    x = np.ascontiguousarray(np.ones(n, np.uint32) * n)
    y = np.ascontiguousarray(np.ones(n, np.uint32) * n, np.uint32)
    u = np.ascontiguousarray(np.zeros(n, np.float64))
    
    # # # # # # # #
    #
    # Column reduction
    #
    # # # # # # # #
    #
    # For a given j, find the i with the minimum cost.
    #
    order = np.lexsort((-i, costs, j))
    min_idx = order[j_index]
    min_i = i[min_idx]
    #
    # v[j] is assigned to the minimum cost over all i
    #
    v = np.ascontiguousarray(costs[min_idx], np.float64)
    #
    # Find the last j for which i was min_i.
    #
    x[min_i] = np.arange(n).astype(np.uint32)
    y[x[x != n]] = np.arange(n).astype(np.uint32)[x != n]
    #
    # Three cases for i:
    #
    # i is not the minimum of any j - i goes on free list
    # i is the minimum of one j - v[j] remains the same and y[x[j]] = i
    # i is the minimum of more than one j, perform reduction transfer
    #
    assignment_count = np.bincount(min_i[min_i != n])
    assignment_count = np.hstack(
        (assignment_count, np.zeros(n - len(assignment_count), int)))
    free_i = assignment_count == 0
    one_i = assignment_count == 1
    #order = np.lexsort((costs, i)) Replace with this after all is done
    order = np.lexsort((j,i))
    j = np.ascontiguousarray(j[order], np.uint32)
    costs = np.ascontiguousarray(costs[order], np.float64)
    i_index = np.ascontiguousarray(i_index, np.uint32)
    i_count = np.ascontiguousarray(i_count, np.uint32)
    if np.any(one_i): 
        reduction_transfer(
            np.ascontiguousarray(np.argwhere(one_i).flatten(), np.uint32),
            j, i_index, i_count, x, u, v, costs)
    #
    # Perform augmenting row reduction on unassigned i
    #
    ii = np.ascontiguousarray(np.argwhere(free_i).flatten(), np.uint32)
    if len(ii) > 0:
        for iii in range (augmenting_row_reductions):
            ii = augmenting_row_reduction(
                n, ii, j, i_index, i_count, x, y, u, v, costs)
    augment(n, ii,
            j, i_index, i_count, x, y, u, v, costs)
    if wants_dual_variables:
        return x,y,u,v
    else:
        return x,y
    
def slow_reduction_transfer(ii, j, idx, count, x, u, v, c):
    '''Perform the reduction transfer step from the Jonker-Volgenant algorithm
    
    The data is input in a ragged array in terms of "i" structured as a
    vector of values for each i,j combination where:
    
    ii - the i to be reduced
    j - the j-index of every entry
    idx - the index of the first entry for each i
    count - the number of entries for each i
    x - the assignment of j to i
    u - the dual variable "u" which will be updated. It should be
        initialized to zero for the first reduction transfer.
    v - the dual variable "v" which will be reduced in-place
    c - the cost for each entry.
    
    The code described in the paper is:
    
    for each assigned row i do
    begin
       j1:=x[i]; u=min {c[i,j]-v[j] | j=1..n, j != j1};
       v[j1]:=v[j1]-(u-u[i]);
       u[i] = u;
    end;
    
    The authors note that reduction transfer can be applied in later stages
    of the algorithm but does not seem to provide a substantial benefit
    in speed.
    '''
    for i in ii:
        j1 = x[i]
        jj = j[idx[i]:(idx[i]+count[i])]
        uu = np.min((c[idx[i]:(idx[i]+count[i])] - v[jj])[jj != j1])
        v[j1] = v[j1] - uu + u[i]
        u[i] = uu
        
def slow_augmenting_row_reduction(n, ii, jj, idx, count, x, y, u, v, c):
    '''Perform the augmenting row reduction step from the Jonker-Volgenaut algorithm
    
    n - the number of i and j in the linear assignment problem
    ii - the unassigned i
    jj - the j-index of every entry in c
    idx - the index of the first entry for each i
    count - the number of entries for each i
    x - the assignment of j to i
    y - the assignment of i to j
    u - the dual variable "u" which will be updated. It should be
        initialized to zero for the first reduction transfer.
    v - the dual variable "v" which will be reduced in-place
    c - the cost for each entry.
    
    returns the new unassigned i
    '''
        
    #######################################
    #
    # From Jonker:
    #
    # procedure AUGMENTING ROW REDUCTION;
    # begin
    # LIST: = {all unassigned rows};
    # for all i in LIST do
    #    repeat
    #    ul:=min {c[i,j]-v[j] for j=l ...n};
    #    select j1 with c [i,j 1] - v[j 1] = u1;
    #    u2:=min {c[i,j]-v[j] for j=l ...n,j< >jl} ;
    #    select j2 with c [i,j2] - v [j2] = u2 and j2 < >j 1 ;
    #    u[i]:=u2;
    #    if ul <u2 then v[jl]:=v[jl]-(u2-ul)
    #    else if jl is assigned then jl : =j2;
    #    k:=y [jl]; if k>0 then x [k]:=0; x[i]:=jl; y [ j l ] : = i ; i:=k
    #  until ul =u2 (* no reduction transfer *) or k=0 i~* augmentation *)
    #  end
    ii = list(ii)
    k = 0
    limit = len(ii)
    free = []
    while k < limit:
        i = ii[k]
        k += 1
        j = jj[idx[i]:(idx[i] + count[i])]
        uu = c[idx[i]:(idx[i] + count[i])] - v[j]
        order = np.lexsort([uu])
        u1, u2 = uu[order[:2]]
        j1,j2 = j[order[:2]]
        i1 = y[j1]
        if u1 < u2:
            v[j1] = v[j1] - u2 + u1
        elif i1 != n:
            j1 = j2
            i1 = y[j1]
        if i1 != n:
            if u1 < u2:
                k -= 1
                ii[k] = i1
            else:
                free.append(i1)
        x[i] = j1
        y[j1] = i
    return np.array(free,np.uint32)

def slow_augment(n, ii, jj, idx, count, x, y, u, v, c):
    '''Perform the augmentation step to assign unassigned i and j
    
    n - the # of i and j, also the marker of unassigned x and y
    ii - the unassigned i
    jj - the ragged arrays of j for each i
    idx - the index of the first j for each i
    count - the number of j for each i
    x - the assignments of j for each i
    y - the assignments of i for each j
    u,v - the dual variables
    c - the costs
    '''
    
    ##################################################
    #
    # Augment procedure: from the Jonker paper.
    #
    # Note:
    #    cred [i,j] = c [i,j] - u [i] - v[j]
    #
    # procedure AUGMENT;
    # begin
    #   for all unassigned i* do
    #   begin
    #     for j:= 1 ... n do 
    #       begin d[j] := c[i*,j] - v[j] ; pred[j] := i* end;
    #     READY: = { ) ; SCAN: = { } ; TODO: = { 1 ... n} ;
    #     repeat
    #       if SCAN = { } then
    #       begin
    #         u = min {d[j] for j in TODO} ; 
    #         SCAN: = {j | d[j] = u} ;
    #         TODO: = TODO - SCAN;
    #         for j in SCAN do if y[j]==0 then go to augment
    #       end;
    #       select any j* in SCAN;
    #       i := y[j*]; SCAN := SCAN - {j*} ; READY: = READY + {j*} ;
    #       for all j in TODO do if u + cred[i,j] < d[j] then
    #       begin
    #         d[j] := u + cred[i,j]; pred[j] := i;
    #         if d[j] = u then
    #           if y[j] is unassigned then go to augment else
    #           begin SCAN: = SCAN + {j} ; TODO: = TODO - {j} end
    #       end
    #    until false; (* repeat always ends with go to augment *)
    #augment:
    #   (* price updating *)
    #   for k in READY do v[k]: = v[k] + d[k] - u;
    #   (* augmentation *)
    #   repeat
    #     i: = pred[j]; y[ j ] := i ; k:=j; j:=x[i]; x[i]:= k
    #   until i = i*
    #  end
    #end
    inf = np.sum(c) + 1
    d = np.zeros(n)
    cc = np.zeros((n,n))
    cc[:,:] = inf
    for i in range(n):
        cc[i,jj[idx[i]:(idx[i]+count[i])]] = c[idx[i]:(idx[i]+count[i])]
    c = cc
    for i in ii:
        print "Processing i=%d" % i
        j = jj[idx[i]:(idx[i] + count[i])]
        d = c[i,:] - v
        pred = np.ones(n, int) * i
        on_deck = []
        ready = []
        scan = []
        to_do = list(range(n))
        try:
            while True:
                print "Evaluating i=%d, n_scan = %d" % (i, len(scan))
                if len(scan) == 0:
                    ready += on_deck
                    on_deck = []
                    umin = np.min([d[jjj] for jjj in to_do])
                    print "umin = %f" % umin
                    scan = [jjj for jjj in to_do if d[jjj] == umin]
                    to_do = [jjj for jjj in to_do if d[jjj] != umin]
                    for j1 in scan:
                        if y[j1] == n:
                            raise StopIteration()
                j1 = scan[0]
                iii = y[j1]
                print "Consider replacing i=%d, j=%d" % (iii, j1)
                scan = scan[1:]
                on_deck += [j1]
                u1 = c[iii, j1] - v[j1] - umin
                for j1 in list(to_do):
                    h = c[iii, j1] - v[j1] - u1
                    print "Consider j=%d as replacement, c[%d,%d]=%f,v[%d]=%f,h=%f, d[j]= %f" % (j1,iii,j1,c[iii,j1],j1,v[j1],h,d[j1])
                    if h < d[j1]:
                        print "Add to chain"
                        pred[j1] = iii
                        if h == umin:
                            if y[j1] == n:
                                raise StopIteration()
                            print "Add to scan"
                            scan += [j1]
                            to_do.remove(j1)
                        d[j1] = h

        except StopIteration:
            # Augment
            print "Augmenting %d" % j1
            for k in ready:
                temp = v[k]
                v[k] = v[k] + d[k] - umin
                print "v[%d] %f -> %f" % (k, temp, v[k])
            while True:
                iii = pred[j1]
                print "y[%d] %d -> %d" % (j1, y[j1], iii)
                y[j1] = iii
                j1, x[iii] = x[iii], j1
                if iii == i:
                    break
    #
    # Re-establish slackness since we didn't pay attention to u
    #
    for i in range(n):
        j = x[i]
        u[i] = c[i,j] - v[j]



import numpy

def outline(labels):
    """Given a label matrix, return a matrix of the outlines of the labeled objects
    
    If a pixel is not zero and has at least one neighbor with a different
    value, then it is part of the outline.
    """
    
    output = numpy.zeros(labels.shape, labels.dtype)
    lr_different = labels[1:,:]!=labels[:-1,:]
    ud_different = labels[:,1:]!=labels[:,:-1]
    d1_different = labels[1:,1:]!=labels[:-1,:-1]
    d2_different = labels[1:,:-1]!=labels[:-1,1:]
    different = numpy.zeros(labels.shape, bool)
    different[1:,:][lr_different]  = True
    different[:-1,:][lr_different] = True
    different[:,1:][ud_different]  = True
    different[:,:-1][ud_different] = True
    different[1:,1:][d1_different] = True
    different[:-1,:-1][d1_different] = True
    different[1:,:-1][d2_different] = True
    different[:-1,1:][d2_different] = True
    #
    # Labels on edges need outlines
    #
    different[0,:] = True
    different[:,0] = True
    different[-1,:] = True
    different[:,-1] = True
    
    output[different] = labels[different]
    return output
    



import numpy as np

def mode(a):
    '''Compute the mode of an array
    
    a: an array
    
    returns a vector of values which are the most frequent (more than one
    if there is a tie).
    '''
    a = np.asanyarray(a)
    if a.size == 0:
        return np.zeros(0, a.dtype)
    aa = a.flatten()
    aa.sort()
    indices = np.hstack([[0], np.where(aa[:-1] != aa[1:])[0]+1, [aa.size]])
    counts = indices[1:] - indices[:-1]
    best_indices = indices[:-1][counts == np.max(counts)]
    return aa[best_indices]



"""FastEMD - python wrapper of the FastEMD algorithm

CellProfiler is distributed under the GNU General Public License,
but this file is licensed under the more permissive BSD license.
See the accompanying file LICENSE for details.

Copyright (c) 2003-2009 Massachusetts Institute of Technology
Copyright (c) 2009-2015 Broad Institute
All rights reserved.

Please see the AUTHORS file for credits.

Website: http://www.cellprofiler.org

FastEMD is a library generously contributed to the open-source community
under a BSD license by it's author, Ofir Pele. Please see the c++ header files
for their copyright. The following papers should be used when citing this
library's implementation of the earth mover's distance:

A Linear Time Histogram Metric for Improved SIFT Matching
 Ofir Pele, Michael Werman
 ECCV 2008
bibTex:
@INPROCEEDINGS{Pele-eccv2008,
author = {Ofir Pele and Michael Werman},
title = {A Linear Time Histogram Metric for Improved SIFT Matching},
booktitle = {ECCV},
year = {2008}
}
 Fast and Robust Earth Mover's Distances
 Ofir Pele, Michael Werman
 ICCV 2009
@INPROCEEDINGS{Pele-iccv2009,
author = {Ofir Pele and Michael Werman},
title = {Fast and Robust Earth Mover's Distances},
booktitle = {ICCV},
year = {2009}
}
"""
from _fastemd import EMD_NO_FLOW, EMD_WITHOUT_TRANSHIPMENT_FLOW,\
     EMD_WITHOUT_EXTRA_MASS_FLOW, emd_hat_int32